#  Copyright (c) 2024. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.

import numpy as np
from scipy.optimize import minimize

from tencirchem import UCCSD
from tencirchem.molecule import h4
from tencirchem.utils.optimizer import soap


def test_optimizer():
    ucc = UCCSD(h4)
    ucc.kernel()

    opt_res = minimize(ucc.energy, ucc.init_guess, method=soap)
    assert np.allclose(opt_res.fun, ucc.e_ucc)



#  Copyright (c) 2024. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.

import numpy as np

from scipy.optimize import OptimizeResult


def soap(fun, x0, args=(), maxfev=2000, callback=None, **kwargs):
    """
    Scipy Optimizer interface for sequantial optimization with
    approximate parabola (SOAP)

    Parameters
    ----------
    fun : callable ``f(x, *args)``
        Function to be optimized.
    x0 : ndarray, shape (n,)
        Initial guess. Array of real elements of size (n,),
        where 'n' is the number of independent variables.
    args : tuple, optional
        Extra arguments passed to the objective function.
    maxfev : int
        Maximum number of function evaluations to perform.
        Default: 2000.
    callback : callable, optional
        Called after each iteration.

    Returns
    -------
    res : OptimizeResult
        The optimization result represented as a SciPy ``OptimizeResult`` object.
        Important attributes are: ``x`` the solution array.
    """

    nfev = 0
    nit = 0

    def _fun(_x):
        nonlocal nfev
        nfev += 1
        return fun(_x, *args)

    trajectory = [x0.copy()]
    vec_list = []
    # direction order
    metric = np.abs(x0)
    for i in np.argsort(metric)[::-1]:
        vec = np.zeros_like(x0)
        vec[i] = 1
        vec_list.append(vec)
    vec_list_copy = vec_list.copy()

    e_list = [_fun(trajectory[-1])]
    offset_list = []
    diff_list = []

    scale = 0.1

    while nfev < maxfev:
        if len(vec_list) != 0:
            vec = vec_list[0]
            vec_list = vec_list[1:]
        else:
            vec_list = vec_list_copy.copy()
            # continue
            if len(trajectory) < len(vec_list_copy):
                continue
            p0 = trajectory[-1 - len(vec_list_copy)]
            f0 = e_list[-1 - len(vec_list_copy)]
            pn = trajectory[-1]
            fn = e_list[-1]
            fe = _fun(2 * pn - p0)
            if fe > f0:
                continue
            average_direction = pn - p0
            if np.allclose(average_direction, 0):
                continue
            average_direction /= np.linalg.norm(average_direction)
            replace_idx = np.argmax(np.abs(diff_list[-len(vec_list_copy) :]))
            df = np.abs(diff_list[-len(vec_list_copy) :][replace_idx])
            if 2 * (f0 - 2 * fn + fe) * (f0 - fn - df) ** 2 > (f0 - fe) ** 2 * df:
                continue
            del vec_list[replace_idx]
            vec_list = [average_direction] + vec_list
            vec_list_copy = vec_list.copy()
            continue

        vec_normed = vec / np.linalg.norm(vec)
        x = [-scale, 0, scale]
        es = [None, e_list[-1], None]
        for j in [0, -1]:
            es[j] = _fun(trajectory[-1] + x[j] * vec_normed)
        if np.argmin(es) == 0:
            x = [-scale * 4, -scale, 0, scale]
            es = [None, es[0], es[1], es[2]]
            for j in [0]:
                es[j] = _fun(trajectory[-1] + x[j] * vec_normed)
        elif np.argmin(es) == 2:
            x = [-scale, 0, scale, scale * 4]
            es = [es[0], es[1], es[2], None]
            for j in [-1]:
                es[j] = _fun(trajectory[-1] + x[j] * vec_normed)
        else:
            assert np.argmin(es) == 1
        a, b, c = np.polyfit(x, es, 2)
        if np.argmin(es) not in [0, 3]:
            offset = b / 2 / a
            e = c - b**2 / 4 / a
        else:
            # print(a, b)
            offset = -x[np.argmin(es)]
            e = np.argmin(es)
        offset_list.append(offset)
        trajectory.append(trajectory[-1] - offset * vec_normed)
        if len(es) == 3:
            e_list.append(e)
        else:
            e_list.append(_fun(trajectory[-1]))
        diff_list.append(e_list[-1] - e_list[-2])

        if callback is not None:
            callback(np.copy(x0))

        nit += 1

    return OptimizeResult(fun=e_list[-1], x=trajectory[-1], nit=nit, nfev=nfev, success=True)



import numpy as np
import pytest

from tencirchem import clear_cache, UCCSD
from tencirchem.molecule import h4, ch4, benzene


def test_clear_cache():
    uccsd = UCCSD(h4)
    e1 = uccsd.kernel()
    clear_cache()
    e2 = uccsd.kernel()
    np.testing.assert_allclose(e2, e1)


@pytest.mark.parametrize("mol", [ch4, benzene])
def test_molecule(mol):
    mol = mol()
    hf = mol.HF()
    hf.kernel()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="tencirchem",
    version="2024.11",
    author="TenCirChem Authors",
    author_email="liw31@163.com",
    description="Efficient quantum computational chemistry based on TensorCircuit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tencent-quantum-lab/tencirchem",
    packages=setuptools.find_packages(),
    python_requires=">=3.9",
    include_package_data=True,
    install_requires=[
        "numpy==1.26.4",
        "scipy==1.13.1",
        "pandas==2.2.3",
        "tensorcircuit[qiskit]==0.12.0",
        "pyscf==2.7.0",
        "openfermion==1.6.1",
        "pylatexenc==2.10",
        "noisyopt==0.2.2",
        "renormalizer==0.0.10",
    ],
    extras_require={
        "jax": ["jax", "jaxlib"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.

"""
Library for variational basis state encoder
https://arxiv.org/abs/2301.01442
"""
import numpy as np
from opt_einsum import contract


def get_psi_indices(dof_nature, b_dof_pidx, n_qubit_per_mode):
    b_dof_vidx = []
    for i in range(max(dof_nature) + 1):
        b_dof_vidx.append(np.where(dof_nature == i)[0])

    # indices used for tensor contraction
    psi_idx = []
    for i in range(max(dof_nature) + 1 + list(dof_nature).count(-1)):
        if i not in b_dof_pidx:
            psi_idx.extend([f"p-{i}"])
        else:
            psi_idx.extend([f"v-{i}-{j}" for j in range(n_qubit_per_mode)])
    psi_idx_top = [f"{i}-top" for i in psi_idx]
    psi_idx_bottom = [f"{i}-bottom" for i in psi_idx]

    return psi_idx_top, psi_idx_bottom, b_dof_vidx


def get_contracted_mpo(h_mpo, b_array, n_qubit_per_mode, b_dof_pidx, psi_indices):
    nbas = b_array.shape[-1]
    b_shape = tuple([2] * n_qubit_per_mode + [nbas])
    args = []
    for i in range(len(h_mpo)):
        args.append(h_mpo[i])
        args.append([f"mpo-{i}", f"p-{i}-top", f"p-{i}-bottom", f"mpo-{i + 1}"])
    for i in range(len(b_array)):
        dof_pidx = b_dof_pidx[i]
        args.append(b_array[i].reshape(b_shape).conj())
        args.append([f"v-{dof_pidx}-{j}-top" for j in range(n_qubit_per_mode)] + [f"p-{dof_pidx}-top"])
        args.append(b_array[i].reshape(b_shape))
        args.append([f"v-{dof_pidx}-{j}-bottom" for j in range(n_qubit_per_mode)] + [f"p-{dof_pidx}-bottom"])
    out = psi_indices.copy()
    out.extend(["mpo-0", f"mpo-{len(h_mpo)}"])
    args.append(out)
    size = round(2 ** (len(psi_indices) // 2))
    h_contracted = contract(*args).reshape(size, size)
    return h_contracted


def get_contract_args(psi, h_mpo, b_array, i, n_qubit_per_mode, psi_idx_top, psi_idx_bottom, b_dof_pidx):
    nbas = b_array.shape[-1]
    psi_shape2 = [2] * len(psi_idx_top)
    b_shape = tuple([2] * n_qubit_per_mode + [nbas])
    # psi
    args = [psi.reshape(psi_shape2).conj(), psi_idx_top] + [psi.reshape(psi_shape2), psi_idx_bottom]
    # H
    for j in range(len(h_mpo)):
        args.append(h_mpo[j])
        args.append([f"mpo-{j}", f"p-{j}-top", f"p-{j}-bottom", f"mpo-{j + 1}"])
    # b
    for j in range(len(b_array)):
        if j == i:
            continue
        k = b_dof_pidx[j]
        indices = [f"v-{k}-{l}-bottom" for l in range(n_qubit_per_mode)] + [f"p-{k}-bottom"]
        args.extend([b_array[j].reshape(b_shape), indices])
        indices = [f"v-{k}-{l}-top" for l in range(n_qubit_per_mode)] + [f"p-{k}-top"]
        args.extend([b_array[j].reshape(b_shape).conj(), indices])
    return args
